/****************************************************************************
**
*W  smith_form.C          Homology source                  Dumas & Saunders
**
*H  @(#)$Id: smith_form.C,v 0.0 1999/08
**
*Y  Copyright (C)  1999,  Us guys
*T Time-stamp: <16 Apr 03 13:53:28 Jean-Guillaume.Dumas@imag.fr> 
**
*/

#include <stdio.h>
#include "matrix.h"
#include "commentator.h"
#include "givinteger.h"

#include "LinBox/lin_spv_bb.h" 
#include "LinBox/lin_wrap_c++.h"

#ifndef GIVABS
#define GIVABS(a) ((a)>0?(a):-(a))
#endif

typedef SparseBlackBoxDom< OperatorWrapper<long> > SPBBlong;

// Methods
// #define hmValenceElim   3
// #define hmValenceBB     4

// Modes
// #define SMITHFORM       0
// #define INTEGRANK       1
// #define TORSIONRANK     2
// #define PRIMERANK       4
// #define PRIMEPOWERRANK  5
// #define CARACTERISTICROWS 6

int main(int argc, char ** argv) {

        // command line :
        // Filename, Hom level, Tim level, Method, Mode, coeff.
	// or
	// Filename, Hom level, Tim level, Method, Mode, coeff, numerator, denominator



    if ((argc != 7) && (argc != 9)) {
        std::cerr << "Wrong number of command-line arguments (" << (argc-1) << ")." << std::endl;
        return 1;
    }

    Invariants<unsigned long long> SmithForm;
    char * File_Name,  * UT, * mat_file = argv[1];
    int is_gzipped = 0;
    size_t s = strlen(mat_file);
    if ((mat_file[--s] == 'z') && (mat_file[--s] == 'g') && (mat_file[--s] == '.')) {
        is_gzipped = 1;
        File_Name = tempnam("/tmp","bbx_");
        UT = new char[s+34+strlen(File_Name)];
        sprintf(UT,"gunzip -c %s > %s", mat_file, File_Name);
        system(UT);
        sprintf(UT,"\\rm %s", File_Name);
    } else
        File_Name = mat_file;
    


    OperatorWrapper<long> LongDom;
    SPBBlong TheMat( LongDom );
    unsigned long ni, nj;
    TheMat.read(ni,nj, File_Name);
    

    int Method =  atoi(argv[4]);
    int Mode =  atoi(argv[5]);

    Commentator GlobComm = Commentator(atoi(argv[3]),atoi(argv[2]));
    Integer R( argv[6] );
    

    MatrixSmithFormValenceCPP(GlobComm, ni, nj, File_Name, GIVABS(Method), Mode, R, SmithForm, argv[7], argv[8] );

    if (Method > 0) {
    	std::cout << "[" << SmithForm.n_row() << ", " << SmithForm.n_col() ;
    	for(TCount i=0; i<SmithForm.size(); ++i) {
        	std::cout << ", [" << SmithForm.getvalue(i) << ", " << SmithForm.getexponent(i) << "]";
    	}
    	std::cout << "]";
    } else {
        if (Mode  == CARACTERISTICROWS) {
            for(unsigned long i = 0; i < SmithForm.size(); ++i) 
                std::cout << SmithForm.getvalue( i ) << " ";
            std::cout << std::endl << SmithForm.rank() << std::endl;
        } else
            std::cout << SmithForm.getexponent(0) << std::endl;
    }


    if (is_gzipped) {
        system(UT);
        delete [] UT;
    }


    return 0;
}
